#include "RepoBuffet.h"
#include <QtCore/QRegExp>
#include "stringmacros.h"

#define S1_NAMEKEY_PREFIX      "/buffet?"
#define S1_NAME_PREFIX         "="
#define S1_PASSKEY_PREFIX      "&"
#define S1_PASS_PREFIX         "="
#define S1_PATH_PREFIX         "&path="

#define S1_NAMEKEY_PREFIX_REGEXP  "/buffet\\?"
#define S1_NAME_PREFIX_REGEXP     "="
#define S1_PASSKEY_PREFIX_REGEXP  "&"
#define S1_PASS_PREFIX_REGEXP     "="
#define S1_PATH_PREFIX_REGEXP     "&path="

#define S1_ARG_REGEXP          "[^&\\s]*"

#define S2_NAMEKEY_PREFIX      "/rhel/"
#define S2_NAME_PREFIX         "="
#define S2_PASSKEY_PREFIX      "/"
#define S2_PASS_PREFIX         "="
#define S2_PATH_PREFIX         "/"

#define S2_NAMEKEY_PREFIX_REGEXP  "/rhel/"
#define S2_NAME_PREFIX_REGEXP     "="
#define S2_PASSKEY_PREFIX_REGEXP  "/"
#define S2_PASS_PREFIX_REGEXP     "="
#define S2_PATH_PREFIX_REGEXP     "/"

#define S2_ARG_REGEXP          "[^/\\s]*"

static inline QChar  toHex(int  c)
{
   if (0 <= c && c <= 9) {
      return QChar('0' + c);
   } else if (10 <= c && c <= 15) {
      return QChar('a' + c - 10);
   } else {
      return QChar('?');
   }
}

static inline int  fromHex(QChar  ch)
{
   int  c = ch.unicode();

   if ('0' <= c && c <= '9') {
      return c - '0';
   } else if ('a' <= c && c <= 'f') {
      return c - 'a' + 10;
   } else if ('A' <= c && c <= 'F') {
      return c - 'A' + 10;
   } else {
      return -1;
   }
}

static QString  httpUnprotect (QString  s)
{
   QRegExp  hexCode("%[0-9a-fA-F][0-9a-fA-F]");

   int  ix = 0;
   for (;;) {
      ix = s.indexOf(hexCode, ix);
      if (ix == -1) break;
      
      int  value = fromHex(s[ix+1])*16 + fromHex(s[ix+2]);

      s.replace(ix, 3, (QChar)value);

      ix++;
   }
   
   return s;
}

static QString  httpProtect (QString  s)
{
   QRegExp  special("["
                    "$&+,/:;=?@"       // http reserved characters
                    " \"<>#%{}|\\\\^~" // http "unsafe" characters
                    "]"
                    "|\\[|\\]");       // http "unsafe" characters

   int  ix = 0;
   for (;;) {
      ix = s.indexOf(special, ix);
      if (ix == -1) break;
      
      QString  hexCode("%");
      hexCode.append(toHex(s[ix].unicode() / 16));
      hexCode.append(toHex(s[ix].unicode() % 16));

      s.replace(ix, 1, hexCode);

      ix += 3;
   }
   
   return s;
}

bool  Repo::Buffet::getAuth (const QString&  url,
                             QString&        nameKey,
                             QString&        name,
                             QString&        passwordKey,
                             QString&        password)
{
   int  nameKeyPrefixIndex;
   int  nameKeyIndex;
   int  namePrefixIndex;
   int  nameIndex;
   int  passKeyPrefixIndex;
   int  passKeyIndex;
   int  passPrefixIndex;
   int  passIndex;
   int  pathPrefixIndex;

   nameKeyPrefixIndex = url.indexOf(S1_NAMEKEY_PREFIX);
   if (nameKeyPrefixIndex != -1) {
      nameKeyIndex       = nameKeyPrefixIndex + STRLEN(S1_NAMEKEY_PREFIX);

      namePrefixIndex    = url.indexOf(S1_NAME_PREFIX, nameKeyIndex);
      if (namePrefixIndex == -1) return false;
      nameIndex          = namePrefixIndex + STRLEN(S1_NAME_PREFIX);

      passKeyPrefixIndex = url.indexOf(S1_PASSKEY_PREFIX, nameIndex);
      if (passKeyPrefixIndex == -1) return false;
      passKeyIndex       = passKeyPrefixIndex + STRLEN(S1_PASSKEY_PREFIX);

      passPrefixIndex    = url.indexOf(S1_PASS_PREFIX, passKeyIndex);
      if (passPrefixIndex == -1) return false;
      passIndex          = passPrefixIndex + STRLEN(S1_PASS_PREFIX);

      pathPrefixIndex    = url.indexOf(S1_PATH_PREFIX, passIndex);
      if (pathPrefixIndex == -1) return false;
   } else {
      nameKeyPrefixIndex = url.indexOf(S2_NAMEKEY_PREFIX);
      if (nameKeyPrefixIndex == -1) return false;
      nameKeyIndex       = nameKeyPrefixIndex + STRLEN(S2_NAMEKEY_PREFIX);

      namePrefixIndex    = url.indexOf(S2_NAME_PREFIX, nameKeyIndex);
      if (namePrefixIndex == -1) return false;
      nameIndex          = namePrefixIndex + STRLEN(S2_NAME_PREFIX);

      passKeyPrefixIndex = url.indexOf(S2_PASSKEY_PREFIX, nameIndex);
      if (passKeyPrefixIndex == -1) return false;
      passKeyIndex       = passKeyPrefixIndex + STRLEN(S2_PASSKEY_PREFIX);

      passPrefixIndex    = url.indexOf(S2_PASS_PREFIX, passKeyIndex);
      if (passPrefixIndex == -1) return false;
      passIndex          = passPrefixIndex + STRLEN(S2_PASS_PREFIX);

      pathPrefixIndex    = url.indexOf(S2_PATH_PREFIX, passIndex);
      if (pathPrefixIndex == -1) return false;
   }

   nameKey     = url.mid(nameKeyIndex, namePrefixIndex    - nameKeyIndex);
   name        = url.mid(nameIndex,    passKeyPrefixIndex - nameIndex);
   passwordKey = url.mid(passKeyIndex, passPrefixIndex    - passKeyIndex);
   password    = url.mid(passIndex,    pathPrefixIndex    - passIndex);

   nameKey     = httpUnprotect(nameKey);
   name        = httpUnprotect(name);
   passwordKey = httpUnprotect(passwordKey);
   password    = httpUnprotect(password);

   return true;
}

bool  Repo::Buffet::getAuth (const Repo::Entry&  entry,
                             QString&            nameKey,
                             QString&            name,
                             QString&            passwordKey,
                             QString&            password)
{
   // I considered having this check for multiple urls that used buffet and
   // had different passwords, and returning false in that case to disable
   // the authentication fields.  But I worried that that would just be
   // irritating to the user, so I didn't do it.

   return (getAuth(entry.urls, 
                   nameKey, name, passwordKey, password)
           || getAuth(entry.mirrorListUrl, 
                      nameKey, name, passwordKey, password));
}

QString  Repo::Buffet::replaceAuth (QString         str,
                                    const QString&  nameKey,
                                    const QString&  name,
                                    const QString&  passwordKey,
                                    const QString&  password)
{
   QRegExp  search1(S1_NAMEKEY_PREFIX_REGEXP S1_ARG_REGEXP
                    S1_NAME_PREFIX_REGEXP    S1_ARG_REGEXP
                    S1_PASSKEY_PREFIX_REGEXP S1_ARG_REGEXP
                    S1_PASS_PREFIX_REGEXP    S1_ARG_REGEXP
                    S1_PATH_PREFIX_REGEXP);
   QString  replace1;
   replace1.append(S1_NAMEKEY_PREFIX).append(httpProtect(nameKey))
           .append(S1_NAME_PREFIX)   .append(httpProtect(name))
           .append(S1_PASSKEY_PREFIX).append(httpProtect(passwordKey))
           .append(S1_PASS_PREFIX)   .append(httpProtect(password))
           .append(S1_PATH_PREFIX);

   QRegExp  search2(S2_NAMEKEY_PREFIX_REGEXP S2_ARG_REGEXP
                    S2_NAME_PREFIX_REGEXP    S2_ARG_REGEXP
                    S2_PASSKEY_PREFIX_REGEXP S2_ARG_REGEXP
                    S2_PASS_PREFIX_REGEXP    S2_ARG_REGEXP
                    S2_PATH_PREFIX_REGEXP);
   QString  replace2;
   replace2.append(S2_NAMEKEY_PREFIX).append(httpProtect(nameKey))
           .append(S2_NAME_PREFIX)   .append(httpProtect(name))
           .append(S2_PASSKEY_PREFIX).append(httpProtect(passwordKey))
           .append(S2_PASS_PREFIX)   .append(httpProtect(password))
           .append(S2_PATH_PREFIX);

   return str.replace(search1, replace1).replace(search2, replace2);
}

QString  Repo::Buffet::replaceAuth (QString         str,
                                    const QString&  name,
                                    const QString&  password)
{
   QString  nameKey;
   QString  nameDummy;
   QString  passwordKey;
   QString  passwordDummy;

   getAuth(str, nameKey, nameDummy, passwordKey, passwordDummy);

   return replaceAuth(str, nameKey, name, passwordKey, password);
}

QString  Repo::Buffet::replaceAuth (QString         str,
                                    const QString&  password)
{
   QString  nameKey;
   QString  name;
   QString  passwordKey;
   QString  passwordDummy;

   getAuth(str, nameKey, name, passwordKey, passwordDummy);

   return replaceAuth(str, nameKey, name, passwordKey, password);
}

QRegExp  Repo::Buffet::oneUrlRegExp ()
{
   return QRegExp("^.*"
                  "(?:"
                  S1_NAMEKEY_PREFIX_REGEXP S1_ARG_REGEXP
                  S1_NAME_PREFIX_REGEXP    S1_ARG_REGEXP
                  S1_PASSKEY_PREFIX_REGEXP S1_ARG_REGEXP
                  S1_PASS_PREFIX_REGEXP    "(" S1_ARG_REGEXP ")"
                  S1_PATH_PREFIX_REGEXP
                  "|"
                  S2_NAMEKEY_PREFIX_REGEXP S2_ARG_REGEXP
                  S2_NAME_PREFIX_REGEXP    S2_ARG_REGEXP
                  S2_PASSKEY_PREFIX_REGEXP S2_ARG_REGEXP
                  S2_PASS_PREFIX_REGEXP    "(" S2_ARG_REGEXP ")"
                  S2_PATH_PREFIX_REGEXP
                  ")"
                  ".*$");
}

QRegExp  Repo::Buffet::urlsRegExp ()
{
   return QRegExp("(?:"
                  S1_NAMEKEY_PREFIX_REGEXP S1_ARG_REGEXP
                  S1_NAME_PREFIX_REGEXP    S1_ARG_REGEXP
                  S1_PASSKEY_PREFIX_REGEXP S1_ARG_REGEXP
                  S1_PASS_PREFIX_REGEXP    "(" S1_ARG_REGEXP ")"
                  S1_PATH_PREFIX_REGEXP
                  "|"
                  S2_NAMEKEY_PREFIX_REGEXP S2_ARG_REGEXP
                  S2_NAME_PREFIX_REGEXP    S2_ARG_REGEXP
                  S2_PASSKEY_PREFIX_REGEXP S2_ARG_REGEXP
                  S2_PASS_PREFIX_REGEXP    "(" S2_ARG_REGEXP ")"
                  S2_PATH_PREFIX_REGEXP
                  ")"
                  "[\\s]*\\b");
}
